package ru.dbotthepony.mc.otm.block.entity.tech

import net.minecraft.core.BlockPos
import net.minecraft.core.Direction
import net.minecraft.world.entity.player.Inventory
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.Level
import net.minecraft.world.level.block.state.BlockState
import net.minecraftforge.common.capabilities.ForgeCapabilities
import net.minecraftforge.energy.IEnergyStorage
import ru.dbotthepony.mc.otm.block.IDroppableContainer
import ru.dbotthepony.mc.otm.block.entity.MatteryDeviceBlockEntity
import ru.dbotthepony.mc.otm.capability.*
import ru.dbotthepony.mc.otm.capability.energy.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.container.MatteryContainer
import ru.dbotthepony.mc.otm.container.HandlerFilter
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.core.math.RelativeSide
import ru.dbotthepony.mc.otm.core.math.facingOne
import ru.dbotthepony.mc.otm.core.util.BESubscribeList
import ru.dbotthepony.mc.otm.menu.tech.BatteryBankMenu
import ru.dbotthepony.mc.otm.registry.MBlockEntities

class BatteryBankBlockEntity(p_155229_: BlockPos, p_155230_: BlockState) : MatteryDeviceBlockEntity(MBlockEntities.BATTERY_BANK, p_155229_, p_155230_), IDroppableContainer {
	var gaugeLevel by synchronizer.float()
		private set

	// 6 на 2
	val container: MatteryContainer = object : MatteryContainer(this::setChanged, CAPACITY) {
		override fun setChanged(slot: Int, new: ItemStack, old: ItemStack) {
			super.setChanged(slot, new, old)
			batteryStatus[slot].value = new.getCapability(ForgeCapabilities.ENERGY).isPresent
			gaugeLevel = (energy.batteryLevel / energy.maxBatteryLevel).toFloat()
		}
	}

	val batteryStatus = ImmutableList(CAPACITY) {
		synchronizer.bool(false, name = "battery$it")
	}

	override val droppableContainer by ::container
	private val itemHandler = container.handler(
		object : HandlerFilter {
			override fun canInsert(slot: Int, stack: ItemStack): Boolean {
				return stack.getCapability(ForgeCapabilities.ENERGY).isPresent
			}

			override fun canExtract(slot: Int, amount: Int, stack: ItemStack): Boolean {
				return true
			}
		}
	)

	init {
		exposeItemsGlobally(itemHandler)
		savetable(::container, INVENTORY_KEY)
	}

	private data class BatteryBankDistribution(val distribution: Array<Decimal>, val maxThroughput: Decimal)

	private inner class BatteryBankEnergy(override val energyFlow: FlowDirection) : IMatteryEnergyStorage {
		override fun extractEnergy(howMuch: Decimal, simulate: Boolean): Decimal {
			return distributeEnergy(isReceiving = false, howMuch, simulate)
		}

		fun getDistribution(isReceiving: Boolean): BatteryBankDistribution {
			val distribution = Array(container.containerSize) { Decimal.ZERO }
			var summ = Decimal.ZERO

			for (i in 0 until container.containerSize) {
				val stack = container.getItem(i)

				if (!stack.isEmpty) {
					stack.energy?.let {
						val diff: Decimal

						if (isReceiving) {
							diff = it.receiveEnergy(it.maxEnergyStoredMattery, true)
						} else {
							diff = it.extractEnergy(it.energyStoredMattery, true)
						}

						distribution[i] = diff
						summ += distribution[i]
					}
				}
			}

			if (!summ.isZero) {
				for (i in 0 until container.containerSize) {
					distribution[i] = distribution[i] / summ
				}
			}

			return BatteryBankDistribution(distribution, summ)
		}

		private fun distributeEnergy(isReceiving: Boolean, howMuch: Decimal, simulate: Boolean, set: Boolean = false): Decimal {
			if (!howMuch.isPositive)
				return Decimal.ZERO

			val distribution = getDistribution(isReceiving)

			if (distribution.maxThroughput.isZero)
				return Decimal.ZERO

			val distList = distribution.distribution
			var summ = Decimal.ZERO

			for (i in 0 until container.containerSize) {
				if (!distList[i].isZero) {
					val stack = container.getItem(i)

					if (!stack.isEmpty) {
						stack.energy?.let {
							val diff: Decimal

							if (isReceiving) {
								diff = it.receiveEnergy(howMuch * distList[i], simulate)
							} else {
								diff = it.extractEnergy(howMuch * distList[i], simulate)
							}

							summ += diff
						}
					}
				}
			}

			if (!simulate && !summ.isZero) {
				setChangedLight()
				gaugeLevel = (batteryLevel / maxBatteryLevel).toFloat()
			}

			return summ
		}

		override fun receiveEnergy(howMuch: Decimal, simulate: Boolean): Decimal {
			return distributeEnergy(isReceiving = true, howMuch, simulate)
		}

		override val canSetBatteryLevel: Boolean
			get() = false

		override var batteryLevel: Decimal
			get() {
				var result = Decimal.ZERO

				for (i in 0 until container.containerSize) {
					val stack = container.getItem(i)

					if (!stack.isEmpty) {
						stack.energy?.let {
							result += it.energyStoredMattery
						}
					}
				}

				return result
			}
			set(value) {
				throw UnsupportedOperationException("Can't set battery value for battery bank")
			}

		override val maxBatteryLevel: Decimal
			get() {
			var result = Decimal.ZERO

			for (i in 0 until container.containerSize) {
				val stack = container.getItem(i)

				if (!stack.isEmpty) {
					stack.energy?.let {
						result += it.maxEnergyStoredMattery
					}
				}
			}

			return result
		}
	}

	private val energyReceiver = BatteryBankEnergy(FlowDirection.INPUT)
	private val energyExtractor = BatteryBankEnergy(FlowDirection.OUTPUT)
	private val energy = BatteryBankEnergy(FlowDirection.BI_DIRECTIONAL)

	override fun createMenu(containerID: Int, inventory: Inventory, ply: Player): AbstractContainerMenu {
		return BatteryBankMenu(containerID, inventory, this)
	}

	init {
		exposeEnergySideless(energy)
		exposeEnergy(RelativeSide.FRONT, energyExtractor)

		for (side in RelativeSide.values()) {
			if (side != RelativeSide.FRONT) {
				exposeEnergy(side, energyReceiver)
			}
		}
	}

	override fun setLevel(p_155231_: Level) {
		super.setLevel(p_155231_)
		checkSurroundings()
	}

	private val consumers = object : BESubscribeList<IEnergyStorage>(this@BatteryBankBlockEntity, ForgeCapabilities.ENERGY) {
		override fun test(t: Direction): Boolean {
			return blockState.facingOne == t
		}
	}

	fun checkSurroundings() = consumers.update(blockState.facingOne::equals)

	fun tick() {
		if (redstoneControl.isBlockedByRedstone)
			return

		for (it in consumers) {
			val (_, maxThroughput) = energy.getDistribution(false)

			if (maxThroughput.isZero)
				continue

			val diff = it.receiveEnergy(maxThroughput, true)

			if (!diff.isZero) {
				val newExtract = energy.extractEnergy(diff, true)
				val newReceive = it.receiveEnergy(newExtract, true)

				val extracted = energy.extractEnergy(newReceive, false)
				val received = it.receiveEnergy(extracted, false)

				if (received < extracted) {
					energy.receiveEnergy(extracted - received, false)
				}
			}
		}
	}

	companion object {
		const val CAPACITY = 6 * 2
	}
}
